# Plan
Here's a checklist of my immediate plans for the future. If you have an idea for a feature, feel free to make an issue.

## v1.2.0
### Initial polishes.
* [x] Refactor module structure.
	* [x] Move many modules handling rendering into `playing/view/`. (Can't do all as visibility rules would need to change.)
		* [x] Stars.
		* [x] Ammo status.
		* [x] Scores renderer.
		* [x] Camera.
		* [x] Particle.
		* [x] Death effect (if possible).
		* [x] Health bar.
	* [x] Move many things into the `world` module.
		* [x] Player.
		* [x] Bullet.
		* [x] Ammo crate.
		* [x] Powerup.
		* [x] Rename `border` to `border_renderer`.
	* [x] Move `texture.rs` to `utils/`.
	* [x] Move `utils/net*` to `net/`.
	* [x] Move `title`, `play_selection`, `settings` and `about` to `gui/`.
	* [ ] ~Move blocks into the world if possible.~

		Can't be bothered right now.
* [x] Prevent players and powerups spawning too close to each other, to avoid an effect being applied for no apparent reason (if a powerup spawns directly on top of a player) which can be confusing (especially if it's a teleportation).

### Plan
* [x] Plan what to do.

### Chat and commands
All commands I'm planning on will be completely determined by the server for now.
* [x] Design a protocol for sending chat messages and commands to the server. Want to completely separate chat from commands by having chat opened by pressing 'T' with commands instead being '/'.
* [x] Implement the GUI on the client side.
	* [x] Get started with recording text input but not adding it to the UI.
	* [x] Prevent moving the player while the text input is open.
	* [x] Disable the text input when pressing enter.
	* [x] Prevent the first 't' or '/' from being added to the line.
	* [x] Get backspace working and rebind PlayingLeave to prevent accidents.
	* [x] Get a very simple GUI of the current line shown.
	* [x] Show some history as well.
* [x] Get networking of messages working.
	* [x] Have the client send out chat messages.
	* [x] Have the server replay chat messages to all clients, including the sender.
	* [x] Receive chat messages and add them to the history. Would need to change the history format.
	* [x] Implement the protocol for commands but without any commands.
* [x] Have the server send a chat message in singleplayer explaining things, including how to leave.
* [x] Improve styling of the GUI.
	* [x] Prevent first-frame jitter on enter.
	* [x] Colour messages.
	* [x] Style and use a consistent font size for the current line.
	* [x] Add a black background with some transparency to make reading text easier.
	* [x] Increase the opacity of this background when opening the text input.
	* [x] Animate changes in background opacity.
	* [x] Make sure the chat isn't too wide.
		* [x] Start it.
		* [x] Finish things by limiting the window height.
	* [x] Test things with DPI scaling.
	* [x] Fade out old messages but still keep them in the chat to read when opening the text input.
	* [ ] ~Slowly decrease the old height to the new height (when decreasing due to messages fading out).~
	* [x] Look into making the background colour be transparent dark grey instead of transparent black.

		All black works best, but completely transparent when closed.
	* [x] Don't show the ">" or "%" if there's no chat history and no text currently in the line.
* [x] Create server command for the admin to message all users either on a single server or all servers.
	* [x] Get started.
	* [x] Fix weird issue where rapidly resizing a terminal running the server causes the server to lock up.
		* [x] Start switching to `reedline`.
		* [x] Polish things up with `reedline`.
		* [x] Get Vi mode working again.
	* [x] Finish things.
* [x] Fix weird issue of what seems to be the ammo text being selected.
	It's instead the window edge being coloured due to it being resized from initial clicking events.
* [x] Use a smaller chat height when the input is closed.
* [x] Clear chat when leaving the game.
* [x] Prevent keys getting "stuck down" when going into chat.
* [x] Show the chat prompt when there's no history and it's open.
* [x] Fix bug with decorative stars when toggling between very wide/narrow and fullscreen.

	Generally happens when both the aspect ratio and window size rapidly change.
* [x] Ensure that FP error doesn't cause the rendered health to be slightly less than zero.
* [x] Add rate limiting to chat to prevent extreme spam.
* [x] Think about whether newlines should be in messages and possibly prevent them.

	Eh, it's fine. The consequences aren't too big if modified clients add newlines to sent messages and I can fix things later.
* [x] Avoid sending messages with just whitespace or empty.
* [x] Prevent disconnection if the chat/command message to-be-sent is too large.

### Simple goal: First to N kills
The exact value of N can be specified by the config.
* [x] Update scores when player is disconnected after violating the protocol by sending a message too long.
* [x] Fix sync bug by ensuring that no "double dipping" happens when the world updates are sent out with the InitWorld message.
* [x] Ensure that sanity check messages are sent after
* [ ] ~Investigate `reedline` and terminal output to stdout vs stderr.~
* [x] Switch from `reedline` to `linefeed`.
* [x] Fix Ctrl+C problem.
* [x] Have the ammo status show the smooth world's ammo count.
* [x] Fix bug where the player can join without a score and take over another player if they disconnect at the bottom code (player still exists in world but not in the clients).
* [x] Fix clippy issue with `Server::update` being too large.
* [x] Refactor code even more to prepare for the goal of first to N kills.
	* [x] Get started.
	* [x] Get things compiling.
	* [x] Make multiple changes to the scores atomic to avoid possible visual artefacts.
* [x] Test things.
* [x] Sort scores by order.
* [x] Implement this goal.
	* [x] Get basic server-side configuration for the scoring and goal.
	* [x] Create a `ClearItems` world event to remove all bullets, ammo crates and powerups from the world, as well as `PlayerReset`.

		This will be emitted whenever the game restarts.
	* [x] Reset the game when the goal is met.
	* [x] Add a server message.
* [x] Fix bug where a player dying/resetting on an effect zone doesn't reset their effect.
* [x] Allow configuring an initial server text message sent to all clients who join the game. This can be instructions of what the scoring and goal is.
* [x] Replace the current singleplayer initial chat message with this.

### Spectator mode
Will be useful for further goals planned in the future, like giving players a limited number of lives and the winner being the last player alive. When a player loses all their lives, they can enter spectator mode.
* [x] Check if things are fine when the server sends an admin message as soon as a player joins (the other shouldn't be first, a violation of the protocol).

	Not a problem, the messages sent before InitWorld are in the misc stream so things are fine.
* [x] Basic spectator mode.
	* [x] Change SERP Play request data to support requesting spectator mode.
	* [x] Prepare the server to support spectating.
	* [x] Get spectating working on the server-side.
		* [x] Modify `NewPlayer` event to take a player id.
	* [x] Get spectating working on the client-side. That is, allow players to join as spectators to which they won't see the HUD.
	* [x] Avoid sending update messages that will be ignored (to save bandwidth).
	* [x] Warn on invalid acks given in `SmoothWorld::sync` method.
	* [x] Fix ack bug.
	* [x] Write comments explaining all of the ack stuff so I can know how things work in the future.
	* [x] Make spectator mode smoother.
	* [x] Allow the player to move the camera while spectating.
		* [x] Keyboard.
		* [x] Dragging.
			* [x] Add action to generalise how it can be done (by default it will be bound to the left mouse button).
			* [x] Get it working.
			* [x] Change cursor when dragging.
* [x] Fix [this](https://youtu.be/LSNQuFEDOyQ) problem in the game.
	* [x] Create decay code.
	* [x] Replace existing code with it.
* [x] Allow requesting spectator mode in the GUI/CLI.
	* [x] Flag in the CLI arguments for the `play` subcommand.
	* [x] Checkbox in the GUI.
		* [x] Local servers.
		* [x] Enter address.
		* [x] Singleplayer.
* [x] Add command to join and leave spectator mode.
	* [x] Create test commands on the server for `spectate` and `play`.
	* [x] Add `help` command as well.
	* [x] Implement these commands.
	* [x] Fix the client to support this.
	* [x] Fix bug with when player joins as spectating then runs `play`.
	* [x] Test that rapidly switching between playing and spectating works fine.
	* [x] Immediately move the camera to the player when they join.
* [x] Update the `status` and LAN discovery to report the total number of clients, including spectators.
* [x] Add option to allow following a player while in spectator mode.
	Here, players would press a key (say 'F' for follow) to toggle if they're following a player or freely, and then while following press the left and right movement keys to cycle between the players. At this point, I might show the HUD of that player.

	* [x] Create actions.
	* [x] Toggle between a player and free movement.
	* [x] Allow moving between each player.
	* [x] Remember the player id when leaving the following state. This is so it can be restored from later.
	* [x] Jump to the player when they leave spectating mode after following another player.
* [x] Privacy improvements.
	One possible problem is the two people are chatting on a public server with no other players in the game, but there might be a spectator eavesdropping them. This can be mitigated by including a status bar below the chat input displaying some information, including the number of players and spectators.

	* [ ] ~Add a server message when players join/leave the game, including spectators.~

		Can make the chat noisy and the status line is good enough.
	* [x] Status line.
		* [x] Create a very basic status line.
		* [x] Show the player and spectator count.
			* [x] Add player count.
			* [x] Have the server send to clients the number of spectators.
			* [x] Add spectator count.
		* [x] Add the ping to the status line, for useful debugging information.
	* [x] Always show the status line, even if the chat should be hidden.
	* [x] Decrease the speed that the ping is updated to avoid it being a distraction.
	* [x] Send a constant player update while spectating to avoid the server getting an idea of what the spectating player is doing.

### Config changes
~This will be allowing config files to be split into multiple parts, one for the world (including special areas, portals and effect zones), another for events, another for LAN discovery and one finally for the goal. The config file can specify through an enum whether each part should be embedded or instead loaded from a file. I can do this recursively if I want, though I probably won't need to.~

### Lobby system
The purpose of this is to have all players be ready when the game starts, with players entering a command like `ready` to the server to indicate that. This lobby can be in the world which probably is the easiest option.

This will result in two states managed by the server that are cycled between: The lobby state and the active playing state. When the server starts, the lobby state is entered. When sufficiently many players are in the lobby and after enough time (or enough players are ready), the active playing state is transitioned into. Then the game continues to play either forever or until a certain goal is reached. After the goal is reached, the lobby state is transitioned into with a message congratulating who won, and the whole cycle repeats.

For "sandbox" games (like right now with no goal, just a kill count for scores), the lobby is unnecessary so this should be something I can configure.

The lobby will also be necessary for some games where players *cannot* join later during the game, like a survival goal where players have a limited number of lives.

* [x] Create a `Lobby` struct containing the time left while in the lobby, and implement a basic lobby with it.
* [x] Allow configuring the lobby time, with 0 indicating the lobby's disabled.
* [x] Have the scores be cleared when in the lobby.
* [x] Avoid showing the "Scores" text when there are no scores.
* [x] Enter the lobby state and reset the time if there are no players in the world (spectators not counted). If the lobby state's disabled, just reset the game.
* [x] Allow configuring of whether the game should restart when there are no players.
* [x] Allow the game to start early through the `ready` command.
	* [x] Get it working at first if all players run `ready`.
	* [x] Loosen the requirement to instead be ≥N% of players ready.
	* [x] Add a command response for when the player's already voted.
	* [x] Print when players indicate they're ready, including the number of players currently ready and the number required for the game to start.
* [x] Create timer.
	* [x] Plan protocol.
	* [x] Implement it on the server to display the amount of time left in the lobby before the game starts.
	* [x] Display the time to the users.
	* [x] Implement all custom formats.
	* [x] Display a shifted time.
	* [x] Prevent jitter from having the timer go back in time.
	* [x] Improve rounding.
	* [x] Improve the formats.
	* [x] Make the timer paused when there the lobby is empty.
	* [x] Ensure that a timer message doesn't come before any `InitWorld` message.
* [x] Have the players spawn around a single point in the lobby, will make things more fun and social.
* [x] Remove the `WorldRng` from move generation, makes things more complicated.
* [x] Instantly change health bar on a `PlayerReset` event.
* [x] Clear all particles on a `ClearItems` event.

### More goals
Want to have many more goals than just "first to N kills", with a way to fine tune things in the config.

* [x] Goals.
	* [x] FirstTo.
		* [x] First to N.
		* [x] Generalised with all comparison operators.
		* [x] Unify goals and avoid translation between them.
	* [x] LastWith.
	* [x] Extremum.
	* [x] Timer.
		* [x] Create the goal.
		* [x] Add a timer to the config so this can be actually used.
		* [x] Fix bug with egui not showing it.
	* [x] Multiple.
	* [x] And and Or (special cases of Multiple).
	* [x] SomeAll.
	* [x] SomeEmpty.
	* [x] None.
* [x] Score controller.
	* [x] Kills.
	* [x] Deaths.
	* [x] Time.
* [x] Reset state of goals.
* [x] Improve scoring configuration.
	* [x] Sorting order.
	* [x] Initial score.
	* [x] Minimum and maximum scores.
	* [x] Whether players can join mid-game.
	* [x] Elimination of players on some condition (like losing all their lives).
	* [x] Fix rare bug discovered as a result of elimination.
	* [x] Check that elimination works.
* [x] Add `RemovePlayer` world event to remove the jitter when a player is eliminated.
	This jitter exists when the player should exist during the update (while physics is being processed with them) but they shouldn't exist at the end of the update due to being eliminated.
* [ ] ~Improve first frame on starting playing.~
	Seems to be fine, just the window size updating.
* [x] Make `r` an alias of `ready`.
* [x] Fix bug where if two players are eliminated at the same time in a game of survival, the game prematurely exits instead of them both winning.
* [x] Occupation game mode.
	The goal of this can be to occupy the opposing player's base for a certain amount of time.
	* [x] Allow block maps to be flipped horizontally or vertically in the server config.
	* [x] Create map for an occupation game.
	* [x] Custom spawning positions.
		* [x] Add custom player spawning positions to the config schema.
		* [x] Change the `MoveChooser` trait to support hinting the type of movement (teleportation vs spawning).
		* [x] Implement custom player spawning.
	* [x] Add an `Occupation` variant to the `ScoringEventListener`.
	* [x] Add a `Pos` variant to the `ScoringEvent`.
	* [x] Get occupation event listeners working.
		* [x] Get a basic occupation game working.
		* [x] Add period.
		* [x] Add capture time.
* [x] Change goal of occupation game.
* [x] Spectator changes.
	* [x] Put all clients out of spectator mode when they enter the lobby. Don't do this after the game ends if the lobby is disabled.
	* [x] Remove jitter when the player leaves spectating mode.
	* [x] Ensure that there is no jitter when the final player eliminated might get put into spectator mode then immediately removed from spectator mode.
	* [x] Ensure that if a player manually enters spectating mode (either with the Play request or by typing `spectate`), they don't re-enter playing mode when switching to the lobby.
	* [x] Ensure that if a player requesting to play while spectating (and if they cannot join) will result in them no longer being manually spectating (so they will join next time).
* [x] Fix bug where a player can occupy the opposing base while in the lobby and have their score increase when the game starts.
	This was a consequence of a sync bug.
* [x] Allow the server config to specify a minimum number of players for the game to start.
	* [x] Add that to `occupation.ron` and `survival.ron`.

### Allow players to voluntarily weaken themselves
This will be similar to SuperTuxKart's handicap system but much more flexible. This will be achieved through a custom server-side command.

This will be called "underpowering".

* [x] Text IO newtype changes.
	* [x] Create newtype to enforce length invariant.
	* [x] Use it.
* [x] Underpowering.
	* [x] Implement functionality in the effect controller.
	* [x] Refactor to share common command code between REPL and text IO.
		* [x] Create generic code.
		* [x] Implement it in REPL.
		* [x] Implement it in user commands.
		* [ ] ~Make the process of getting all help text generic.~
		* [x] Make help command generic and implement it for user commands.
	* [x] Create command to use it.
	* [x] Get the command working.
	* [x] Improve it to allow clearing a specific effect (another way to do `up EFFECT 1`).
	* [x] Update help text.

### Finishing touches
* [x] Improve physics on very high speeds, such as being able to move outside the world.
* [x] Fix bug with inaccurate delta time used for sanity checking and chat rate limiting.
* [x] Add warning on internet connections with security disabled that messages might be eavesdropped.
	* [x] Create `Security` enum and add a method returning it to the `MessageStream` trait.
	* [x] Add this and a warning in the insecure case to the status bar.
* [x] Fix DHCP MiTM.
* [x] Resolve all TODOs.
* [x] Update documentation.
	* [x] Chat and commands, including spectating and underpowering.
	* [x] Settings.
	* [x] Goals, lobby, timer in server config.
	* [x] Explanation about security.
* [x] Change blue player colour and chat size.
* [x] Update text in About state.
* [x] Update SERP version.

## v1.2.1
* [ ] Flatpak.

## v1.3.0
* [ ] Allow the player to choose their colour and name.
* [ ] Teams.
* [ ] Update SERP version.
